""" coordination transformation """

from typing import List, Tuple
import numpy as np

def m2pix(height: float,
          extention: float,
          radius: float,# = 6_378_137 ,
          distorsion_scaling: float = 1.) -> float:
    """ convert from meter in pixel

    Arguments:
        height(float): image height
        extention(float): latitude extention of image in pixel
        radius(float): planet radius in metern
        distorsion_scaling(float): Scaling factor for
                                  distortion between 0. and 1.

    Returns:
        float: conversionsfactor pix/m
    """

    return (180. / np.pi) * height * \
            distorsion_scaling / extention / radius

def km2pix(height: float,
           extention: float,
           radius: float = 6_378.137,
           distorsion_scaling: float = 1.) -> float:
    """ convert from kilometer in pixel

    Arguments:
        height(float): image height
        extention(float): latitude extention of image in pixel
        radius(float): planet radius in kilometer default is
                       the earth radius (6_378.137km)
        distorsion_scaling(float): Scaling factor for
                                  distortion between 0. and 1.

    Returns:
        float: conversionsfactor pix/m
    """

    assert 0.0 < distorsion_scaling <= 1.0, \
        f"apu.geo.Coord: distorsion_scaling {distorsion_scaling} has to" +\
         " be in the interval ]0,1]"

    return m2pix(height=height,
                 extention=extention,
                 radius=radius*1000.0,
                 distorsion_scaling=distorsion_scaling) / 1000.0

def pix2carree(pixel: List[float],
               area: List[Tuple[float]],
               image_size: List[int],
               origin: str = "upper") -> tuple:
    """ Convert image pixel position to Carree lat/long
        ASSAMTION: central median is 0 => (long [-180,180[)

    Arguments:
        pixel(List[float]): (u,v) coordinate in image
        area(List[Tuple[float]]): ((u,v)_min, (u,v)_max) the area in the image
                    to search in
        image_size(List[int]): image size in width and height
        origin(str): image convention. where is the image origin.
                'upper' means the origin [0,0] is in the upper left corner
                'lower' means that the image origin is in the lower left corner

    Returns:
        tuple: (lat, lon) coordinates in a Plate Carree image
    """

    lat = (pixel[0] / image_size[0]) * (area[1][0] - area[0][0]) + area[0][0]
    lon = (pixel[1] / image_size[1]) * (area[1][1] - area[0][1])
    lon = lon + area[0][1] if origin == "lower" else area[0][1] - lon

    return (lat, lon)

# pylint: disable=C0103
def carree2pix(coord: List[float],
               area: List[Tuple[float]],
               image_size: List[int],
               origin: str = "upper") -> tuple:
    """ Convert Carree lat/long to image poxel position

    Arguments:
        coord(List[float]): (lat, lon) coordinate in image
        area(List[Tuple[float]]): ((u,v)_min, (u,v)_max) the area in the image
                    to search in
        image_size(List[int]): image size in width and height
        origin(str): image convention. where is the image origin.
                'upper' means the origin [0,0] is in the upper left corner
                'lower' means that the image origin is in the lower left corner

    Returns:
        tuple: (u, v) pixel coordinate
    """

    u = image_size[0] * (coord[0] - area[0][0]) / (area[1][0] - area[0][0])
    v = image_size[1] / (area[1][1] - area[0][1])
    v = v * (coord[1] - area[0][1]) if origin == "lower" else \
        v * (area[1][1] - coord[1])

    return (u, v)



""" apu.dp: anton python utils design pattern module """

__version__ = (0, 0, 0)
__email__ = "anton.feldmann@gmail.com"
__author__ = "anton feldmann"

from apu.dp.null import Null
from apu.dp.blackboard import Blackboard, MetaInfo

__all__ = ['Null', "Blackboard", "MetaInfo"]



""" apu.datastructure.memorywrapper """

import abc
from typing import Any
import dill as pickle

from apu.dp.null import Null
from apu.exception.readwrongfile import ReadWrongFile
from apu.io.dill import reconstruct, load


class MemoryWrapper:
    """ wrap the memory and setup steps.
        future change to additional wrapper
    """
    def __init__(self, **kwargs):
        """ init the object """
        self._mem = None
        self._config = kwargs
        self.setup()

    @abc.abstractmethod
    def setup(self):
        """ memory setup """

    @abc.abstractmethod
    def close(self):
        """ set the Pointer to Null """
        self._mem = Null()

    @abc.abstractmethod
    def set(self, key: str, value: Any) -> bool:
        """ set a key to the memory value
        Arguments:
            key(str): string typed key
            value(Any): memory object

        Return:
            (bool): true if successfull stored
        """
        return True

    @abc.abstractmethod
    def get(self, key: str) -> Any:
        """ get the memory value by string

        Arguments:
            key(str): pointer to memory object

        Returns:
            (Any): memory object

        """
        return Null()

    @abc.abstractmethod
    def delete(self, key: str) -> bool:
        """ delete the memory object

        Arguments:
            key(str): pointer to memory object

        Returns:
            (bool): true if successfull deleted
        """

        return True

    @abc.abstractmethod
    def has(self, key: str) -> bool:
        """ has pointer to memory?

        Arguments:
            key(str): request key

        Returns:
            (bool): True if the pointer is in memory
        """
        return False

    @abc.abstractmethod
    def _get_all(self) -> dict:
        """ get the whole managed memory
        Returns:
            (dict): get all serialized data in the blackboard
        """
        return dict()

    @abc.abstractmethod
    def _restore(self, kv_pairs):
        """
        Arguments:
            ky_pairs(dict): key value pairs

        Returns:
            (bool): true if the key value pair is successfully stored
        """
        return True

    @staticmethod
    def transform_value_to_pickle(value: Any) -> str:
        """ serialize an object

        Arguments:
            value(Any): the memory

        Returns:
            (str): serialized object
        """
        return reconstruct(value)

    @staticmethod
    def transform_pickle_to_value(data: str) -> Any:
        """ deserialize object

        Arguments:
            data(str): serialized value

        Returns:
            (Any): deserialized Data

        """
        return load(data)

    def save(self, file_path: str) -> bool:
        """ store the memory to file

        Arguments:
            file_path(str): the path to the fail containing the memory

        Returns:
            (bool): True if the storage was successful
        """
        whole_data = self._get_all()
        with open(file_path, 'wb') as outfile:
            pickle.dump(whole_data, outfile, protocol=pickle.HIGHEST_PROTOCOL)
        return True

    def load(self, file_path: str) -> bool:
        """ load the data from file

        Arguments:
            file_path(str): the path to the fail containing the memory

        Returns:
            (bool): True if the storage was successful
        """
        with open(file_path, 'rb') as infile:
            read_data = pickle.load(infile)
        if not isinstance(read_data, dict):
            raise ReadWrongFile(
                f"File contents must be dictionary data: {read_data}")
        self._restore(read_data)
        return True



""" null objects do nothing only marking the position as null"""

class Null:
    """ Null Object implementation

        This class ignores all parameters parsed to the object
        on creation.
        The created instances do nothing and mark an object
        nor nothing
    """
    def __init__(self, *args, **kargs) -> None:
        """ Ignore the function call """
        return None

    def __call__(self, *args, **kwargs):
        """ Ignore the function call """
        return self

    def __getattr__(self, name):
        """ Ignore the function call """
        return self

    def __setattr__(self, name, value):
        """ Ignore the function call """
        return self

    def __delattr__(self, name):
        """ Ignore the function call """
        return self

    def __repr__(self) -> str:
        """ string representation

        Returns:
            (str): the String \"<Null>\"

        Examples:
        ..  example-code:
            >>> n = Null()
            >>> repr(n)
            <Null>

        """
        return "<Null>"

    def __str__(self) -> str:
        """ string representation

        Returns:
            (str): the String \"Null\"

        Examples:
        ..  example-code:
            >>> n = Null()
            >>> str(n)
            Null
        """
        return "Null"



""" this module read the verstion tag number
and set the project version to the git tag version. """

import os
import re
import tempfile
from git import Repo

def replace_line_in_file(fpath, old_line_start, new_line):
    """ find and delete string in a file """
    assert os.path.exists(fpath)
    assert (old_line_start and str(old_line_start))
    assert (new_line and str(new_line))

    replaced = False
    written = False

    try:
        with open(fpath, 'r+') as fhandle:  # open for read/write -- alias to f
            lines = fhandle.readlines()  # get all lines in file

            breaks = True

            for line in lines:
                if old_line_start in line:
                    breaks = False  # line not found in file, do nothing

            if breaks:
                pass
            else:
                tmpf = tempfile.NamedTemporaryFile(
                    delete=True, mode='r+')  # temp file opened for writing

                for line in lines:  # process each line
                    if old_line_start in line:  # find the line we want
                        tmpf.write(new_line)  # replace it
                        replaced = True
                    else:
                        tmpf.write(line)  # write old line unchanged

                tmpf.flush()
                tmpf.seek(0, os.SEEK_SET)

                if replaced:  # overwrite the original file
                    fhandle.seek(0)  # beginning of file
                    fhandle.truncate()  # empties out original file
                    for tmplines in tmpf.readlines():
                        fhandle.write(
                            tmplines)  # writes each line to original file
                        written = True

                tmpf.close()  # tmpfile auto deleted
            fhandle.close()  # we opened it , we close it

    except IOError as ioe:  # if something bad happened.
        print("ERROR", ioe)
        fhandle.close()
        return False

    return replaced and written  # replacement happened with no errors = True


def setversion(repopath, versionfile):
    """ set the version number in a given file """

    repo = Repo.init(repopath, bare=False)
    assert not repo.bare

    try:
        tags = sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
        latest_tag = tags[-1]

        print(f"found tag {latest_tag}")

        match_obj = re.match(r'v(\d+).(\d+).(\d+)', str(latest_tag), re.I)

        major = match_obj.group(1)
        minor = match_obj.group(2)
        patch = match_obj.group(3)

        replace_line_in_file(
            versionfile, "__version__",
            f"__version__ = ({int(major)}, {int(minor)}, {int(patch)})\n")
    except: # pylint: disable=W0702
        print("set the version to 0.1.0 because no tag set")
        replace_line_in_file(versionfile, "__version__",
                             "__version__ = (0, 1, 0)\n")



""" module builds parallel for loops """

from contextlib import closing
import multiprocessing.pool
from typing import Any, Callable, List, Tuple

def parallel_for(
        loop_callback: Callable[[Any], Any],
        parameters: List[Tuple[Any, ...]],
        nb_threads: int = 8,
) -> List[Any]:
    """Execute a for loop body in parallel
    .. note:: Race-Conditions
        Code executation in parallel can cause into an "race-condition"
        error.

    Arguments:
        loop_callback(Callable): function callback running in the
                                             loop body
        parameters(List[Tuple]): element to execute in parallel

    Returns:
        (List[Any]): list of values

    Examples:
    .. example-code::
        >>> x = lambda x: x ** 2
        >>> parallel_for(x, [y for y in range(10)])
        [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

    """
    with closing(multiprocessing.pool.ThreadPool(nb_threads)) as pool:
        return pool.map(loop_callback, parameters)



""" dill pickle related functions """
from typing import Any

import dill as pickle

from apu.exception.unsupporteddatatype import UnsupportedDataType


def reconstruct(value: Any) -> str:
    """serialize the data

    Arguments:
        value(Any): object to serialize

    Returns:
        (str): serialized data

    Raises:
        UnsupportedDataType: the data are not serializeable

    """
    try:
        value = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
    except pickle.PicklingError as pickle_e:
        raise UnsupportedDataType(
            f"Cannot serialize given data: {value}. Details: {pickle_e}"
        ) from pickle_e
    return value


def load(data: str) -> Any:
    """serialize the data

    Arguments:
        data(str): string to deserialize

    Returns:
        (Any): deserialized data

    Raises:
        UnsupportedDataType: the data are not deserializeable

    """
    try:
        value = pickle.loads(data)
    except pickle.UnpicklingError as upe:
        raise UnsupportedDataType(
            f"Cannot deserialize given data: {data}. Details: {upe}") from upe
    return value



""" apu.datetime: anton python utils datetime module """

__version__ = (0, 0, 0)
__email__ = "anton.feldmann@gmail.com"
__author__ = "anton feldmann"

from apu.time.date_time import DateTime

__all__ = ['DateTime']



""" date time related functions """

import datetime as dt
import time
import random
import pytz

class DateTime:
    """Datetime contains the datetime utility functions"""
    @staticmethod
    def datetime2unix(datetime_: dt.datetime) -> int:
        """convert a datetime object into
            a unixtimestamp

        Arguments:
            datetime_(dt.datetime): datetime

        Returns:
            (int): unix timestamp

        Examples:
        ..  example_code::
            >>> from apu.time.date_time import DateTime
            >>> DateTime.unixtime2date(datetime.now())
            1234567890.123

        """

        return time.mktime(datetime_.timetuple())

    @staticmethod
    def unixtime2date(unix: float) -> dt.datetime:
        """convert a uniy timestamp to datetime object

        Arguments:
            unix(float): unix timestamp

        Returns:
            (datetime.datetime): datetime

        Examples:
        ..  example_code::
            >>> from apu.time.date_time import DateTime
            >>> DateTime.unixtime2date(1234567890.123)
            2009-02-13 23:31:30.123000
        """
        assert isinstance(unix, float), "can not convert to \
                                         datetime make sure unix \
                                         is of type int"

        return dt.datetime.utcfromtimestamp(unix)


    @staticmethod
    def add_time(datetime_: dt.datetime,
                 day: int = 0,
                 hour: int = 0,
                 minute: int = 0,
                 seconds: int = 0) -> dt.datetime:
        """ add time to a datetime object with respect to the
            timezone

            This function will keep the correct timezone

        Arguments:
            datetime_(datetime.datetime): datetime object
            day(int): days
            hour(int): hours
            seconds(int): seconds

        Returns:
            (datetime.datetime): time in the correct timezone
        """

        day = abs(day)
        hour = abs(hour)
        minute = abs(minute)
        seconds = abs(seconds)

        # calculate all in seconds
        seconds += minute * 60
        seconds += hour * 60 ** 2
        seconds += day * 24 * 60 ** 2

        # calculate timezone
        time_zone = datetime_ + dt.timezone(seconds=seconds)
        return time_zone.astimezone(pytz.utc).astimezone(time_zone.tzinfo)

    @staticmethod
    def generate(start: dt.datetime,
                 end: dt.datetime,
                 random_callback=random.Random) -> dt.datetime:
        """generate a random date.

        Arguments:
            start(datetime.datetime): start time
            end(datetime.datetime): end time
            random_callback(random.Random): random number generator

        Returns:
            (datetime.datetime): datetime

        Raises:
            ValueError: the start end oder is wrong

        Examples:
        ..  example_code::
            >>> import random
            >>> import datetime as dt
            >>> from apu.time.date_time import DateTime

            >>> rand = random.Random()
            >>> rand.seed(0)

            >>> print(DateTime.generate(start=dt.datetime(2020,1,2),
            >>>       end=dt.datetime(2020,2,3),
            >>>       random_callback=rand))
            2020-01-29 00:30:57.535096
        """
        if start > end:
            raise ValueError(f"{start} should be before {end} !!!")

        time_intervall = end - start
        random_time = time_intervall * random_callback.random()
        return start + random_time



""" directory related functions"""

from typing import Any, List

from apu.datastructures.memorywrapper import MemoryWrapper
from apu.dp.null import Null

class Dictionary:
    """ shared Memory dictionary """

    __SHARED_MEMORY = {}

    def __init__(self) -> None:
        """ init shared memory """
        self._dict = Dictionary.__SHARED_MEMORY

    def set(self, key: str, value: Any) -> bool:
        """ set a key value pair into the dictionary

        Arguments:
            key(str): key value has to be string
            value(Any): any value is allowed

        Returns:
            (bool): successful added
        """
        self._dict[key] = value
        return True

    def get(self, key: str) -> Any:
        """ get an dictionary element by key

        Arguments:
            key(str): key pointing to the
                      memory related to it

        Returns:
            (Any): the value or a Null object
        """
        if key in self._dict:
            return self._dict[key]

        return Null()

    def keys(self) -> List[str]:
        """ get  all keys of the data storage

        Returns:
            (List[str]): return a list of strings
        """
        return self._dict.keys()

    def delete(self, key: str) -> None:
        """ delete a key value pair

        Arguments:
            key(str): key pointing to memory position

        """
        del self._dict[key]

    def exists(self, key: str) -> bool:
        """ check if the key is in the list of
            pointers in the memory

        Arguments:
            key(str): pointer to value

        Returns:
            (bool): true if the key is a valid pointer

        """
        return key in self._dict

    def flush(self):
        """ clean the memory """
        self._dict.clear()

    @property
    def all(self) -> dict:
        """ get the whole dict

        Returns:
            (dict): the shared memory dictionary

        """
        return self._dict

class DictionaryWrapper(MemoryWrapper):
    """Dictionary class wrapper class.
       This is used for using Dictionary as a memory.
    """

    def setup(self):
        """ setup the Wrapper with a dictionary object"""
        self._mem = Dictionary()

    def close(self):
        """ close the memory and delete all items"""
        self._mem.flush()

    def set(self, key: str, value: Any) -> bool:
        """ store data to memory

        Arguments:
            key(str): pointer to memory
            value(Any): memory to store

        Returns:
            (bool): True if the memory are set successfull
        """
        data = MemoryWrapper.transform_value_to_pickle(value)
        self._mem.set(key, data)
        return True

    def get(self, key: str) -> Any:
        """get the memory utilizing a reference

        Arguments:
            key(str): key to memory

        Returns:
            (Any): memory or Null-Object
        """
        data = self._mem.get(key)
        if data:
            value = MemoryWrapper.transform_pickle_to_value(data)
        else:
            value = Null()
        return value

    def delete(self, key: str) -> bool:
        """ delete the value related to the pointer

        Arguments:
            key(str): point to memory

        Returns:
            (bool): successfully deleted
        """
        if key in self._mem.keys():
            self._mem.delete(key)
            return True
        return False

    def has(self, key: str) -> bool:
        """ does points contain the key

        Arguments:
            key(str): pointer wo memory

        Returns:
            (bool): True if the pointers contains the key
        """
        return self._mem.exists(key)

    def _get_all(self) -> dict:
        """ the managed memory

        Returns:
            (dict): the whole (serialized data)
        """
        return self._mem.all

    def _restore(self, kv_pairs: dict) -> bool:
        """ map a directory into the memory

        Arguments:
            kv_pairs(dict): dictionary. key can also be bytes

        Returns:
            (bool): True if succeed to store kv_pairs to memory

        """
        self._mem.flush()
        for key, val in kv_pairs.items():
            if isinstance(key, bytes):
                key = key.decode("utf-8")
            self._mem.set(key, val)
        return True



"""
build function to use for setup.py
"""
import os
import sys
import subprocess

from distutils.command.build_py import build_py_2to3 as _build_py
from distutils.command.clean import clean as _clean

from distutils.spawn import find_executable

def find_protoc():
    """Locates protoc executable"""
    if 'PROTOC' in os.environ and os.path.exists(os.environ['PROTOC']):
        protoc = os.environ['PROTOC']
    else:
        protoc = find_executable('protoc')

    if protoc is None:
        sys.stderr.write(
            'protoc not found. Is protobuf-compiler installed? \n'
            'Alternatively, you can point the PROTOC environment' + \
            ' variable at a local version.')
        sys.exit(1)

    return protoc

class BuildProtoBuf(_build_py):
    """ find and execute protobuf related work """
    def run(self):
        """ run the proto_builder """
        for package in self.packages:
            packagedir = self.get_package_dir(package)

            for protofile in filter(lambda x: x.endswith('.proto'),
                                    os.listdir(packagedir)):
                source = os.path.join(packagedir, protofile)
                output = source.replace('.proto', '_pb2.py')
                print(f"Generating {output}...")

                if (not os.path.exists(output) or \
                    (os.path.getmtime(source) > os.path.getmtime(output))):
                    protoc_command = [find_protoc(), \
                                      "--python_out=.", source]
                    if subprocess.call(protoc_command) != 0:
                        sys.exit(-1)
        _build_py.run(self)

class CleanProtoBuf(_clean):
    """ clean protobuf info """
    def run(self):
        """ Delete generated files in the code tree. """
        for (dirpath, _, filenames) in os.walk("."):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if filepath.endswith("_pb2.py"):
                    os.remove(filepath)
        # _clean is an old-style class, so super() doesn't work.
        _clean.run(self)



""" apu.io anton python utils input output module """

__version__ = (0, 0, 0)
__email__ = "anton.feldmann@gmail.com"
__author__ = "anton feldmann"

from apu.io.dill import reconstruct, load

__all__ = ['reconstruct', "load"]



""" apu.mp: anton python utils multiprocessing module """

__version__ = (0, 0, 0)
__email__ = "anton.feldmann@gmail.com"
__author__ = "anton feldmann"

from apu.mp.for_ import parallel_for

__all__ = ['parallel_for']



""" dictionary key related errors """

class ExistingKey(Exception):
    """ the key allready exists """

class NonExistingKey(Exception):
    """ the key does not exists """

class KeyNotString(Exception):
    """ key type is not String """



""" apu.setup: anton python utils setup module """

__version__ = (0, 0, 0)
__email__ = "anton.feldmann@gmail.com"
__author__ = "anton feldmann"

from apu.setup.module import Module
from apu.setup.protobuf import find_protoc, BuildProtoBuf, CleanProtoBuf

__all__ = ['Module',
           'find_protoc',
           'BuildProtoBuf',
           'CleanProtoBuf']



#!/usr/bin/env python

"""apu: Antons Python Utilities."""

# Third party
from setuptools import setup

requires_datetime = ["pytz"]
requires_setup = ["GitPython"]
requires_geographie = ["numpy"]
requires_all = (
    requires_datetime
    + requires_setup
    + requires_geographie
)

setup(
    version="0.1.3",
    package_data={"apu": []},
    project_urls={
        'Documentation': 'https://afeldman.github.io/apu/',
        'Source': 'https://github.com/afeldman/apu',
        'Tracker': 'https://github.com/afeldman/apu/issues',
    },
    extras_require={
        "all": requires_all,
        "datetime": requires_datetime,
        "setup": requires_setup,
        "geo": requires_geographie,
    },
)


""" check if a module is installed and import the module if
the module exists """
from importlib.util import find_spec, module_from_spec


class ModuleNotImportedError(ModuleNotFoundError):
    """ Extends ModuleNotFoundError. Because a Module can also
    available but not imported """


class Module:
    """in this class all module containing functions are placed."""
    @staticmethod
    def check(module_name: str):
        """ check if a module exists without importing it

        Arguments:
            module_name(str): module name

        Returns:
            (ModuleSpec): Module specification

        Raises:
            ModuleNotFoundError: Module not found
            AttributeError: given attribute error

        """
        try:
            module_spec = find_spec(module_name)
        except ModuleNotFoundError as mnfe:
            raise ModuleNotFoundError(f"{module_name} not found") from mnfe
        except AttributeError as ate:
            raise AttributeError(f"{module_name} attribute not valid") from ate

        if module_spec is not None:
            return module_spec

        raise ModuleNotFoundError(f"{module_name} not found")

    @staticmethod
    def import_module_from_spec(module_name: str):
        """If the module is found utilizing check_module

        Arguments:
            module_name(str): name of the module to import

        Returns:
            (module): imported module

        Raises:
            ModuleNotImportedError: cannot import module
            ModuleNotFoundError: cannot find the module.
                                 install the module first
            AttributeError: given attribute for the module
                            check is not valid
        """
        try:
            module_spec = Module.check(module_name)

            if module_spec:
                module = module_from_spec(module_spec)
                module_spec.loader.exec_module(module)
                return module
            raise ModuleNotImportedError(f"{module_name} not imported!!")
        except ModuleNotFoundError as mnf:
            raise ModuleNotFoundError(f"{module_name} not found") from mnf
        except AttributeError as ate:
            raise AttributeError(f"{module_name} attribute not valid") from ate



""" apu.dp.backboard implementation """

import json
import os

from typing import Callable, Any, List

from apu.datastructures.dictionary import DictionaryWrapper
from apu.exception.dict_key import ExistingKey, NonExistingKey, KeyNotString
from apu.exception.noteditable import NotEditable
from apu.exception.directory import NonExistingDirectory
from apu.exception.io import UnsafeLoading


class MetaInfo:
    """ additional information and data manipulation utilizing callbacks """
    def __init__(self, read_only: bool = True):
        """ are the values read only

        Arguments:
            read_only(bool): set the value data to read_only
        """
        self.read_only = read_only
        self._callback = list()

    def __call__(self, value: Any):
        """ call all callback functions on given value

        Arguments:
            value(Any): value to run with the callbacks
        """
        for call_back in self._callback:
            call_back(value)

    @property
    def callback(self) -> List[Callable[[Any], Any]]:
        """ get the callbacks

        Returns:
            (List[Callable[[Any], Any]]): list of all callbacks
        """
        return self._callback

    @callback.setter
    def callback(self, callback: Callable[[Any], Any]):
        """ add a new callback to the list of callbacks

        Arguments:
            callback(Callable[[Any], Any]): callable structure
        """
        self._callback.append(callback)

    def del_callback(self, callback: Callable[[Any], Any]):
        """ get one callback if the callback is in the list

        Arguments:
            callback(Callable[[Any], Any]): callable structure
        """
        if callback in self._callback:
            self._callback.remove(callback)

    def clean(self):
        """ delete all callbacks """
        del self._callback[:]


class Blackboard:
    """a shared memory structure to handover information
       in different objects"""
    def __init__(self):
        self._memory_wrapper = DictionaryWrapper()
        self._meta_info = {}

    def close(self):
        """ delete the meta information and close the memory"""
        del self._meta_info
        self._memory_wrapper.close()

    def set(self, key: str, value: Any, read_only: bool = False) -> bool:
        """ set the memory related to the reference key

        Arguments:
            key(str): reference
            value(Any): memory value
            read_only(bool): store the data read only?

        Returns:
            (bool): True if the memory could be placed

        Raises:
            KeyNotString: the reference key is not a string
            ExistingKey: the reference key is already set
        """
        if not isinstance(key, str):
            raise KeyNotString(
                f"Blackboard data `key` {key} should be `str` type.")

        if key in self._meta_info:
            raise ExistingKey("Given `key` already exists in blackboard")

        if self._memory_wrapper.set(key, value):
            self._meta_info[key] = MetaInfo(read_only=read_only)
            return True
        return False

    def get(self, key: str) -> Any:
        """ get the memory

        Arguments:
            key(str): reference key

        Returns:
            (Any): memory value

        Raises:
            NonExistingKey: key does not exists
        """
        if key not in self._meta_info:
            raise NonExistingKey
        return self._memory_wrapper.get(key)

    def update(self, key: str, value: Any) -> bool:
        """ change the memory

        Arguments:
            key(str): reference key
            value(Any): Memory value

        Returns:
            (bool): true if the data are places successfully

        Raises:
            NonExistingKey: key does not exists
            NotEditable: the memory is read only
        """
        if key not in self._meta_info:
            raise NonExistingKey
        meta_info = self._meta_info[key]
        if meta_info.read_only:
            raise NotEditable("Cannot update read-only data")

        if self._memory_wrapper.set(key, value):
            meta_info.callback(value)
            return True
        return False

    def drop(self, key: str) -> bool:
        """ drop memory with pointer

        Arguments:
            key(str): pointer to memory

        Returns:
            (bool): Ture if the drop was successful

        Raises:
            NonExistingKey: key to memory is unknown

        """
        if key not in self._meta_info:
            raise NonExistingKey

        if self._memory_wrapper.delete(key):
            meta_info = self._meta_info[key]
            meta_info.clear_callbacks()
            del self._meta_info[key]
            return True
        return False

    def clear(self):
        """ clean the memory for all reference keys"""
        for key in self.keys(in_list=True):
            self.drop(key)

    def keys(self, in_list: bool = False) -> List[str]:
        """ get a list of all reference keys

        Arguments:
            in_list(bool): as a python list

        Returns:
            List[str]: an iterator or python list object

        """
        if in_list:
            return list(self._meta_info.keys())
        return self._meta_info.keys()

    def register_callback(self, key: str,
                          callback: Callable[[Any], Any]) -> int:
        """ register a callback on the memory values

        Arguments:
            key(str): reference key
            callback(Callable[[Any], Any]): callback function

        Returns:
            int: python id Object

        Raises:
            NonExistingKey: the key does not exists
        """
        if key not in self._meta_info:
            raise NonExistingKey
        meta_info = self._meta_info[key]
        meta_info.add_callback(callback)
        return id(callback)

    def remove_callback(self, key: str, callback: Callable[[Any], Any]) -> int:
        """ remove the callback from the memory

        Arguments:
            key(str): reference key
            callback(Callable[[Any], Any]): delete this callback

        Returns:
            int: python object id

        Raises:
            NonExistingKey: key does not exists
        """
        if key not in self._meta_info:
            raise NonExistingKey
        meta_info = self._meta_info[key]
        meta_info.remove_callback(callback)
        return id(callback)

    def clear_callbacks(self, key: str):
        """ clear the callback

        Arguments:
            key(str): reference key

        Raises:
            NonExistingKey: key does not exists
        """
        if key not in self._meta_info:
            raise NonExistingKey
        meta_info = self._meta_info[key]
        meta_info.clear_callbacks()

    def save(self, dir_path: str = './.blackboard'):
        """ save the blackboard

        Arguments:
            dir_path(str): file path
        """
        if not os.path.exists(dir_path):
            os.mkdir(dir_path, 0o755)
        blackboard_file_path = os.path.join(dir_path, '.blackboard.pickle')
        meta_info_file_path = os.path.join(dir_path, '.blackboard.meta')
        self._memory_wrapper.save(blackboard_file_path)
        self._save_meta_info(meta_info_file_path)

    def load(self, dir_path: str = './.blackboard', safe: bool = True):
        """ load the data from blackboard file

        Arguments:
            dir_path(str): filepath
            safe(bool): load safe

        Raises:
            UnsafeLoading: unsafe loading
            NonExistingDirectory: path does not exists
        """
        if self.keys(in_list=True):
            if safe:
                raise UnsafeLoading
            self.clear()
        if os.path.exists(dir_path):
            blackboard_file_path = os.path.join(dir_path, '.blackboard.pickle')
            meta_info_file_path = os.path.join(dir_path, '.blackboard.meta')
            self._memory_wrapper.load(blackboard_file_path)
            self._load_meta_info(meta_info_file_path)
        else:
            raise NonExistingDirectory

    def _save_meta_info(self, file_path: str):
        """store additional meta information in json file

        Arguments:
            file_path(str): file path

        """
        saved_meta_info = {}
        for key, meta_info in self._meta_info.items():
            read_only = meta_info.read_only
            saved_meta_info[key] = read_only
        with open(file_path, 'w') as outfile:
            json.dump(saved_meta_info, outfile)

    def _load_meta_info(self, file_path: str):
        """load the meta information from file

        Arguments:
            file_path: path to meta file
        """
        with open(file_path, 'r') as infile:
            saved_meta_info = json.load(infile)
        if self._meta_info:
            self._meta_info.clear()
        for key, read_only in saved_meta_info.items():
            self._meta_info[key] = MetaInfo(read_only=read_only)
